import re
import shutil
import string
import unicodedata
import uuid
from typing import Callable

from .entities import *


class SNBTStream:
    """SNBT解析流"""

    def __init__(self, snbt: str):
        self.cur = 0
        self.snbt = snbt
        self.len = len(snbt)

    def peek(self, length: int):
        return self.snbt[self.cur:self.cur + length]

    def read(self, length: int = 1):
        if self.cur + length > self.len:
            return ""
        self.cur += length
        return self.snbt[self.cur - length:self.cur]

    def read_until(self, predicate: Callable[[str], bool]):
        buffer = []
        while self.cur < self.len:
            if predicate(self.snbt[self.cur]):
                break
            buffer.append(self.snbt[self.cur])
            self.cur += 1
        return "".join(buffer)


class SNBTParseError(Exception):
    """SNBT解析错误"""

    def __init__(self, message, position, snbt, length=1):
        super().__init__(message),
        self.message = message
        self.position = position
        self.snbt = snbt
        self.length = length
        self._terminal_width = None

    def __str__(self):
        if self._terminal_width is None:
            self._terminal_width = shutil.get_terminal_size((80, 20)).columns
        pos = self.position
        context_length = self._terminal_width - 4  # 留出边距
        half_context = context_length // 2
        disp_str = self.snbt

        if len(self.snbt) > context_length:
            start = max(0, pos - half_context)
            end = min(len(self.snbt), pos + half_context)
            disp_str = self.snbt[start:end]
            if start > 0:
                disp_str = "..." + disp_str
                pos = pos - start + 3
            if end < len(self.snbt):
                disp_str = disp_str + "..."

        indicator = " " * pos + "^" * self.length
        return f"SNBTParseError: {self.message}\n{disp_str}\n{indicator}"


class SNBTParser:
    """SNBT解析器"""

    def __init__(self, snbt: str):
        self.stream = SNBTStream(snbt)
        self.last_type = None

    def parse_number(self):
        end_values = ['}', ']', ',']
        start_pos = self.stream.cur
        number = self.stream.read_until(lambda x: x in end_values)
        number = number.lower()
        unsigned = False
        is_float = False
        if number.startswith("0x") or number.startswith("0b"):
            number = number.replace("_", "")
            num_type = TagId.TAG_INT
            limit = 2147483648
        else:
            if number.endswith('b'):
                num_type = TagId.TAG_BYTE
                number = number[:-1]
                limit = 128
                if number.endswith('u'):
                    unsigned = True
                    number = number[:-1]
                    limit = 256
                elif number.endswith('s'):
                    number = number[:-1]
            elif number.endswith('s'):
                num_type = TagId.TAG_SHORT
                number = number[:-1]
                limit = 2147483648
                if number.endswith('u'):
                    unsigned = True
                    number = number[:-1]
                    limit = 4294967296
                elif number.endswith('s'):
                    number = number[:-1]
            elif number.endswith('l'):
                num_type = TagId.TAG_LONG
                number = number[:-1]
                limit = 9223372036854775808
                if number.endswith('u'):
                    unsigned = True
                    number = number[:-1]
                    limit = 18446744073709551616
                elif number.endswith('s'):
                    number = number[:-1]
            elif number.endswith('i'):
                num_type = TagId.TAG_INT
                number = number[:-1]
                limit = 2147483648
                if number.endswith('u'):
                    unsigned = True
                    number = number[:-1]
                    limit = 4294967296
                elif number.endswith('s'):
                    number = number[:-1]
            elif number.endswith('f'):
                is_float = True
                number = number[:-1]
                num_type = TagId.TAG_FLOAT
                limit = 3.40282346639e+38
            elif number.endswith('d'):
                is_float = True
                number = number[:-1]
                num_type = TagId.TAG_DOUBLE
                limit = 1.7976931348623157e+308
            else:
                if "." in number or "e" in number:
                    is_float = True
                    num_type = TagId.TAG_DOUBLE
                    limit = 1.7976931348623157e+308
                else:
                    num_type = TagId.TAG_INT
                    limit = 2147483648

        try:
            if is_float:
                res = float(number)
            else:
                res = int(number, 0)
        except ValueError:
            raise SNBTParseError(
                "Invalid number: " + number,
                start_pos,
                self.stream.snbt,
                self.stream.cur - start_pos,
            )
        if unsigned:
            if not (0 <= res < limit):
                raise SNBTParseError(
                    f"Number {number} out of range",
                    start_pos,
                    self.stream.snbt,
                    self.stream.cur - start_pos,
                )
            bit_width = {
                TagId.TAG_BYTE: 8,
                TagId.TAG_SHORT: 16,
                TagId.TAG_INT: 32,
                TagId.TAG_LONG: 64,
            }
            max_unsigned = 2 ** bit_width[num_type]
            if res >= max_unsigned // 2:
                res -= max_unsigned
        else:
            if not (-limit <= res < limit):
                raise SNBTParseError(
                    f"Number {number} out of range",
                    start_pos,
                    self.stream.snbt,
                    self.stream.cur - start_pos,
                )

        return get_tag_class(num_type)(res)

    def _parse_escape(self, char):
        """解析转义字符"""
        if char == "x":
            hex_str = self.stream.read(2)
            if len(hex_str) != 2 or not re.fullmatch(r"[0-9a-fA-F]{2}", hex_str):
                raise SNBTParseError("Invalid \\x escape", self.stream.cur - 2, self.stream.snbt, 4)
            return chr(int(hex_str, 16))
        elif char == "u":
            hex_str = self.stream.read(4)
            if len(hex_str) != 4 or not re.fullmatch(r"[0-9a-fA-F]{4}", hex_str):
                raise SNBTParseError("Invalid \\u escape", self.stream.cur - 4, self.stream.snbt, 6)
            return chr(int(hex_str, 16))
        elif char == "U":
            hex_str = self.stream.read(8)
            if len(hex_str) != 8 or not re.fullmatch(r"[0-9a-fA-F]{8}", hex_str):
                raise SNBTParseError("Invalid \\U escape", self.stream.cur - 8, self.stream.snbt, 10)
            return chr(int(hex_str, 16))
        elif char == "N":
            if self.stream.peek(1) != "{":
                raise SNBTParseError("Expected '{' after \\N", self.stream.cur - 1, self.stream.snbt, 2)
            name = self.stream.read_until(lambda x: x == "}")
            try:
                return unicodedata.lookup(name)
            except KeyError:
                raise SNBTParseError(
                    f"Unknown Unicode name: {name}",
                    self.stream.cur - len(name) - 3,
                    self.stream.snbt,
                    len(name) + 4
                )
        else:  # 经典转义（\n, \t, \" 等）
            escapes = {
                "\\": "\\", "\"": "\"", "'": "'",
                "n": "\n", "r": "\r", "t": "\t",
                "b": "\b", "f": "\f", "s": " "
            }
            return escapes.get(char, char)  # 未知转义（如 \z）原样保留

    def _is_escaped(self):
        """检查当前字符是否被转义"""
        if self.stream.cur == 0:
            return False  # 第一个字符不可能被转义

        # 从当前指针向前查找连续的 \，统计数量
        backslash_count = 0
        pos = self.stream.cur - 1
        while pos >= 0 and self.stream.snbt[pos] == "\\":
            backslash_count += 1
            pos -= 1

        # 奇数个 \ 表示当前字符被转义
        return backslash_count % 2 == 1

    def parse_string(self):
        start_pos = self.stream.cur
        self.stream.read_until(lambda x: not x.isspace())

        # 判断是否带引号
        if self.stream.peek(1) in ["\"", "\'"]:
            prefix = self.stream.read()
            end_condition = lambda x: x == prefix
        else:
            prefix = None
            end_condition = lambda x: x in ["}", "]", ",", ":"] and not self._is_escaped()

        res = []
        while True:
            char_pos = self.stream.cur
            char = self.stream.read(1)
            if not char:
                if prefix:
                    raise SNBTParseError(
                        "Unclosed string",
                        start_pos,
                        self.stream.snbt,
                        self.stream.cur - start_pos
                    )
                break

            if char == "\\":  # 转义字符
                next_char = self.stream.read(1)
                if not next_char:
                    raise SNBTParseError("Incomplete escape sequence", char_pos, self.stream.snbt, 2)
                res.append(self._parse_escape(next_char))
            elif end_condition(char):  # 终止符
                if not prefix:
                    self.stream.cur -= 1  # 非引号字符串回退指针
                break
            else:
                res.append(char)

        return TAG_String("".join(res))

    def parse_list(self):
        """列表及数组解析"""

        # 判断是否是数组
        if self.stream.peek(3) == "[I;":
            self.stream.read(3)
            list_type = TagId.TAG_INT_ARRAY
        elif self.stream.peek(3) == "[B;":
            self.stream.read(3)
            list_type = TagId.TAG_BYTE_ARRAY
        elif self.stream.peek(3) == "[L;":
            self.stream.read(3)
            list_type = TagId.TAG_LONG_ARRAY
        else:
            list_type = TagId.TAG_LIST
            self.stream.read()
        value = []
        index = 0  # 数组索引
        while True:
            if self.stream.peek(1) == "]":
                self.stream.read()
                break
            if self.stream.peek(1) == ",":
                self.stream.read()
            else:
                element_start_pos = self.stream.cur  # 记录元素开始位置
                number = self.parse()

                # 检查类型
                if list_type == TagId.TAG_INT_ARRAY and not isinstance(number, TAG_Int):
                    raise SNBTParseError(
                        f"Invalid type at index {index}: expected TAG_Int, got {type(number).__name__}",
                        element_start_pos,
                        self.stream.snbt,
                        self.stream.cur - element_start_pos,
                    )
                elif list_type == TagId.TAG_BYTE_ARRAY and not isinstance(number, TAG_Byte):
                    raise SNBTParseError(
                        f"Invalid type at index {index}: expected TAG_Byte, got {type(number).__name__}",
                        element_start_pos,
                        self.stream.snbt,
                        self.stream.cur - element_start_pos,
                    )
                elif list_type == TagId.TAG_LONG_ARRAY and not isinstance(number, TAG_Long):
                    raise SNBTParseError(
                        f"Invalid type at index {index}: expected TAG_Long, got {type(number).__name__}",
                        element_start_pos,
                        self.stream.snbt,
                        self.stream.cur - element_start_pos,
                    )
                else:
                    value.append(number)
                    index += 1
        if list_type == TagId.TAG_INT_ARRAY:
            return TAG_Int_Array([ele.value for ele in value])
        elif list_type == TagId.TAG_BYTE_ARRAY:
            return TAG_Byte_Array(bytearray([ele.value for ele in value]))
        elif list_type == TagId.TAG_LONG_ARRAY:
            return TAG_Long_Array([ele.value for ele in value])
        else:  # 列表
            types = [type(ele) for ele in value]
            if len(set(types)) == 1:
                return TAG_List(value)
            else:
                value_new = []
                for ele in value:
                    if isinstance(ele, TAG_Compound):
                        value_new.append(ele)
                    else:
                        comp = TAG_Compound()
                        comp[""] = ele
                        value_new.append(comp)
                return TAG_List(value_new)

    def parse_compound(self):
        """解析compound"""
        res = TAG_Compound()
        self.stream.read()
        while True:
            if self.stream.peek(1) == "}":
                self.stream.read()
                break
            if self.stream.peek(1) == ",":
                self.stream.read()
            else:
                key = self.parse_string()
                if not key.value:  # 空键
                    raise SNBTParseError(
                        f"Expected str at {self.stream.cur}",
                        self.stream.cur,
                        self.stream.snbt
                    )
                split = self.stream.read()
                if split != ":":
                    raise SNBTParseError(
                        f"Expected : at {self.stream.cur}",
                        self.stream.cur,
                        self.stream.snbt
                    )
                start_pos = self.stream.cur  # 记录值开始位置
                value = self.parse()
                if isinstance(value, TAG_String):
                    if value.value == "true":
                        value = TAG_Byte(1)
                    elif value.value == "false":
                        value = TAG_Byte(0)
                    elif value.value.startswith("bool("):
                        param_stream = SNBTStream(value.value[5:])
                        param = param_stream.read_until(lambda x: x == ")")
                        param_stream.read()
                        if param_stream.read():
                            raise SNBTParseError(
                                f"Invalid expression: {value.value}",
                                start_pos,
                                self.stream.snbt,
                                self.stream.cur - start_pos
                            )
                        parse_res = self.parse_snbt(param)
                        if parse_res.value:
                            value = TAG_Byte(1)
                        else:
                            value = TAG_Byte(0)
                    elif value.value.startswith("uuid("):
                        param_stream = SNBTStream(value.value[5:])
                        param = param_stream.read_until(lambda x: x == ")")
                        param_stream.read()
                        if param_stream.read():
                            raise SNBTParseError(
                                f"Invalid expression: {value.value}",
                                start_pos,
                                self.stream.snbt,
                                self.stream.cur - start_pos
                            )
                        parse_res = self.parse_snbt(param)
                        if not isinstance(parse_res, TAG_String):
                            raise SNBTParseError(
                                f"Invalid type at {value.value}: expected TAG_String, got {type(parse_res).__name__} ",
                                start_pos,
                                self.stream.snbt,
                                self.stream.cur - start_pos
                            )
                        value = TAG_Int_Array(self.uuid_to_int_array(parse_res.value))

                res[key.value] = value
        return res

    def parse(self):
        self.stream.read_until(lambda x: not x.isspace())
        prefix = self.stream.peek(1)
        if not prefix:
            raise SNBTParseError(
                f"Expected value at {self.stream.cur}",
                self.stream.cur,
                self.stream.snbt
            )
        if prefix == "{":
            return self.parse_compound()
        elif prefix == "[":
            return self.parse_list()
        elif prefix in string.digits + "+-":
            return self.parse_number()
        elif prefix in ["\"", "\'"]:
            return self.parse_string()
        else:
            return self.parse_string()

    @staticmethod
    def uuid_to_int_array(uuid_str):
        """uuid字符串转换"""

        u = uuid.UUID(uuid_str)
        # 获取UUID的128位整数表示
        uuid_int = u.int
        # 将128位整数拆分为4个32位整数
        int_array = [
            (uuid_int >> 96) & 0xFFFFFFFF,
            (uuid_int >> 64) & 0xFFFFFFFF,
            (uuid_int >> 32) & 0xFFFFFFFF,
            uuid_int & 0xFFFFFFFF
        ]
        # 处理符号位（如果需要有符号整数）
        int_array = [x if x < 0x80000000 else x - 0x100000000 for x in int_array]
        return int_array

    @classmethod
    def parse_snbt(cls, snbt: str):
        return cls(snbt).parse()



from enum import Enum
from struct import Struct
from io import BytesIO
from collections.abc import MutableSequence, Sequence, MutableMapping
from typing import Optional, List


class TagId(Enum):
    TAG_END = 0
    TAG_BYTE = 1
    TAG_SHORT = 2
    TAG_INT = 3
    TAG_LONG = 4
    TAG_FLOAT = 5
    TAG_DOUBLE = 6
    TAG_BYTE_ARRAY = 7
    TAG_STRING = 8
    TAG_LIST = 9
    TAG_COMPOUND = 10
    TAG_INT_ARRAY = 11
    TAG_LONG_ARRAY = 12

    @classmethod
    def __missing__(cls, key):
        raise ValueError(f'Unknown tag id: {key}')


class TAG(object):
    id: TagId

    def __init__(self, name=None, value=None):
        self.name = name
        self.value = value

    def to_binary(self):
        raise NotImplementedError()


class TAG_End(TAG):
    id = TagId.TAG_END
    fmt = Struct('<b')

    def __init__(self):
        super().__init__()

    def to_binary(self):
        return self.fmt.pack(0)


class TAG_OneNumber(TAG):
    fmt = None
    snbt_suffix = None

    def __init__(self, value=None, name=None, binary=None):
        super().__init__(name, value)
        if binary:
            self.__parse_binary(binary)

    def __parse_binary(self, binary: BytesIO):
        self.value = self.fmt.unpack(binary.read(self.fmt.size))[0]

    def to_binary(self):
        return self.fmt.pack(self.value)

    def to_snbt(self):
        return str(self.value) + self.snbt_suffix

    def format_string(self, layer: int = 0):
        res = ""
        if layer:
            res += '  ' * layer
        res += f"{self.__class__.__name__}({self.name if self.name else ''}): {str(self.value)}"
        return res

    def __str__(self):
        return self.format_string()

    def __repr__(self):
        return self.format_string()


class TAG_Byte(TAG_OneNumber):
    id = TagId.TAG_BYTE
    fmt = Struct('<b')
    snbt_suffix = 'b'


class TAG_Short(TAG_OneNumber):
    id = TagId.TAG_SHORT
    fmt = Struct('<h')
    snbt_suffix = 's'


class TAG_Int(TAG_OneNumber):
    id = TagId.TAG_INT
    fmt = Struct('<i')
    snbt_suffix = ''


class TAG_Long(TAG_OneNumber):
    id = TagId.TAG_LONG
    fmt = Struct('<q')


class TAG_Float(TAG_OneNumber):
    id = TagId.TAG_FLOAT
    fmt = Struct('<f')
    snbt_suffix = 'f'


class TAG_Double(TAG_OneNumber):
    id = TagId.TAG_DOUBLE
    fmt = Struct('<d')
    snbt_suffix = 'd'


class TAG_Byte_Array(TAG, MutableSequence):
    id = TagId.TAG_BYTE_ARRAY

    def __init__(self, value: Optional[bytearray] = None, name=None, binary=None):
        super().__init__(name, value)
        if binary:
            self.__parse_binary(binary)

    def __parse_binary(self, binary: BytesIO):
        length = TAG_Int(binary=binary).value
        self.value = bytearray(binary.read(length))

    def to_binary(self):
        return TAG_Int(len(self.value)).to_binary() + self.value

    def to_snbt(self):
        res = "[B;"
        for i in self.value:
            res += f"{int(i)}B,"
        if res[-1] == ',':
            res = res[:-1]
        res += "]"
        return res

    def format_string(self, layer: int = 0):
        res = ""
        if layer:
            res += '  ' * layer
        res += f"{self.__class__.__name__}({self.name if self.name else ''}): {str(self.value)}"
        return res

    def __str__(self):
        return self.format_string()

    def __repr__(self):
        return self.format_string()

    def __len__(self):
        return len(self.value)

    def __getitem__(self, index):
        return self.value[index]

    def __setitem__(self, index, value):
        self.value[index] = value

    def __delitem__(self, key):
        del self.value[key]

    def __iter__(self):
        return iter(self.value)

    def insert(self, index, value):
        self.value.insert(index, value)

    def append(self, value):
        self.value.append(value)


class TAG_Int_Array(TAG, MutableSequence):
    id = TagId.TAG_INT_ARRAY

    def __init__(self, value: Optional[List[int]] = None, name=None, binary=None):
        super().__init__(name, value)
        if binary:
            self.__parse_binary(binary)

    def __parse_binary(self, binary: BytesIO):
        length = TAG_Int(binary=binary).value
        self.value = []
        for _ in range(length):
            self.value.append(TAG_Int(binary=binary).value)

    def to_binary(self):
        return TAG_Int(len(self.value)).to_binary() + b''.join(TAG_Int(value).to_binary() for value in self.value)

    def to_snbt(self):
        res = "[I;"
        for i in self.value:
            res += f"{i},"
        if res[-1] == ',':
            res = res[:-1]
        res += "]"
        return res

    def format_string(self, layer: int = 0):
        res = ""
        if layer:
            res += '  ' * layer
        res += f"{self.__class__.__name__}({self.name if self.name else ''}): {str(self.value)}"
        return res

    def __str__(self):
        return self.format_string()

    def __repr__(self):
        return self.format_string()

    def __len__(self):
        return len(self.value)

    def __getitem__(self, index):
        return self.value[index]

    def __setitem__(self, index, value):
        self.value[index] = value

    def __delitem__(self, key):
        del self.value[key]

    def __iter__(self):
        return iter(self.value)

    def insert(self, index, value):
        self.value.insert(index, value)

    def append(self, value):
        self.value.append(value)


class TAG_Long_Array(TAG, MutableSequence):
    id = TagId.TAG_LONG_ARRAY

    def __init__(self, value: Optional[List[int]] = None, name=None, binary=None):
        super().__init__(name, value)
        if binary:
            self.__parse_binary(binary)

    def __parse_binary(self, binary: BytesIO):
        length = TAG_Int(binary=binary).value
        self.value = []
        for _ in range(length):
            self.value.append(TAG_Long(binary=binary).value)

    def to_binary(self):
        return TAG_Int(len(self.value)).to_binary() + b''.join(
            TAG_Long(value=value).to_binary() for value in self.value)

    def to_snbt(self):
        res = "[B;"
        for i in self.value:
            res += f"{i}L,"
        if res[-1] == ',':
            res = res[:-1]
        res += "]"
        return res

    def format_string(self, layer: int = 0):
        res = ""
        if layer:
            res += '  ' * layer
        res += f"{self.__class__.__name__}({self.name if self.name else ''}): {str(self.value)}"
        return res

    def __str__(self):
        return self.format_string()

    def __repr__(self):
        return self.format_string()

    def __len__(self):
        return len(self.value)

    def __getitem__(self, index):
        return self.value[index]

    def __setitem__(self, index, value):
        self.value[index] = value

    def __delitem__(self, key):
        del self.value[key]

    def __iter__(self):
        return iter(self.value)

    def insert(self, index, value):
        self.value.insert(index, value)

    def append(self, value):
        self.value.append(value)


class TAG_String(TAG, Sequence):
    id = TagId.TAG_STRING

    def __init__(self, value: Optional[str] = None, name=None, binary=None):
        super().__init__(name, value)
        if binary:
            self.__parse_binary(binary)

    def __parse_binary(self, binary: BytesIO):
        length = Struct('<H').unpack(binary.read(2))[0]
        self.value = binary.read(length).decode('utf-8')

    def to_binary(self):
        return Struct('<H').pack(len(self.value)) + self.value.encode('utf-8')

    def to_snbt(self):
        return f'"{self.value}"'

    def format_string(self, layer: int = 0):
        res = ""
        if layer:
            res += '  ' * layer
        res += f'{self.__class__.__name__}({self.name if self.name else ""}): "{str(self.value)}"'
        return res

    def __str__(self):
        return self.format_string()

    def __repr__(self):
        return self.format_string()

    def __len__(self):
        return len(self.value)

    def __getitem__(self, index):
        return self.value[index]

    def __iter__(self):
        return iter(self.value)


class TAG_List(TAG, MutableSequence):
    id = TagId.TAG_LIST

    def __init__(self, value=None, name=None, binary=None, tag_id=None):
        super().__init__(name, [])
        if binary:
            self.__parse_binary(binary)
        elif value:
            self.tag_id = tag_id
            for tag in value:
                if not isinstance(tag, TAG):
                    raise TypeError(f"{tag} is not a TAG")
                if self.tag_id is None:
                    self.tag_id = tag.id
                if tag.id != self.tag_id:
                    raise ValueError(f"{tag} is not a {TagId(self.tag_id).name}")
                self.value.append(tag)

    def __parse_binary(self, binary: BytesIO):
        self.tag_id = TagId(Struct('<b').unpack(binary.read(1))[0])
        length = Struct('<i').unpack(binary.read(4))[0]
        for _ in range(length):
            self.value.append(get_tag_class(self.tag_id)(binary=binary))

    def to_binary(self):
        res = Struct('<b').pack(self.tag_id.value) + Struct('<i').pack(len(self.value))
        for tag in self.value:
            if not isinstance(tag, TAG_Compound):
                res += tag.to_binary()
            else:
                res += tag.to_binary(header=False)
        return res

    def to_snbt(self):
        res = "["
        for tag in self.value:
            if self.tag_id == TagId.TAG_COMPOUND and not tag.name:
                res += f"{tag.to_snbt()[2:-1]},"
            else:
                res += f"{tag.to_snbt()},"
        if res[-1] == ',':
            res = res[:-1]
        res += "]"
        return res

    def format_string(self, layer: int = 0):
        res = ""
        if layer:
            res += '  ' * layer
        res += f"{self.__class__.__name__}({self.name}): {len(self.value)} entries [\n"
        for tag in self.value:
            res += f"{tag.format_string(layer + 1)}\n"
        if layer:
            res += '  ' * layer
        res += "]"
        return res

    def __str__(self):
        return self.format_string()

    def __repr__(self):
        return self.format_string()

    def __len__(self):
        return len(self.value)

    def __getitem__(self, index):
        return self.value[index]

    def __setitem__(self, index, value: TAG):
        if not isinstance(value, TAG):
            raise TypeError("value must be a TAG")
        if value.id != self.tag_id:
            raise TypeError(f"value must be a {get_tag_class(self.tag_id).__name__}")
        self.value[index] = value

    def __delitem__(self, key):
        del self.value[key]

    def __iter__(self):
        return iter(self.value)

    def insert(self, index, value: TAG):
        if not isinstance(value, TAG):
            raise TypeError("value must be a TAG")
        if value.id != self.tag_id:
            raise TypeError(f"value must be a {get_tag_class(self.tag_id).__name__}")
        self.value.insert(index, value)

    def append(self, value: TAG):
        if not isinstance(value, TAG):
            raise TypeError("value must be a TAG")
        if value.id != self.tag_id:
            raise TypeError(f"value must be a {get_tag_class(self.tag_id).__name__}")
        self.value.append(value)


class TAG_Compound(TAG, MutableMapping):
    id = TagId.TAG_COMPOUND

    def __init__(self, name=None, binary=None):
        super().__init__(name, [])
        if binary:
            self.__parse_binary(binary)

    def __parse_binary(self, binary: BytesIO):
        while True:
            tag_id = TagId(Struct('<b').unpack(binary.read(1))[0])
            if tag_id == TagId.TAG_END:
                break
            tag_name = TAG_String(binary=binary).value
            new_tag = get_tag_class(tag_id)(name=tag_name, binary=binary)
            self.value.append(new_tag)

    def to_binary(self, header=True):
        if header:
            res = Struct('<b').pack(self.id.value)
            if self.name:
                res += Struct('<H').pack(len(self.name)) + self.name.encode('utf-8')
            else:
                res += Struct('<H').pack(0)
            for tag in self.value:
                # print(tag, tag.id, tag.name)
                res += Struct('<b').pack(tag.id.value) + Struct('<H').pack(len(tag.name)) + tag.name.encode('utf-8')
                if not isinstance(tag, TAG_Compound):
                    res += tag.to_binary()
                else:
                    res += tag.to_binary(header=False)
            return res + TAG_End().to_binary()
        else:
            res = b''
            for tag in self.value:
                res += Struct('<b').pack(tag.id.value) + Struct('<H').pack(len(tag.name)) + tag.name.encode('utf-8')
                if not isinstance(tag, TAG_Compound):
                    res += tag.to_binary()
                else:
                    res += tag.to_binary(header=False)
            return res + TAG_End().to_binary()

    def to_snbt(self):
        res = "{"
        for tag in self.value:
            res += f"{tag.name}:{tag.to_snbt()},"
        if res[-1] == ',':
            res = res[:-1]
        res += "}"
        return res

    def format_string(self, layer: int = 0):
        res = ""
        if layer:
            res += '  ' * layer
        res += f"{self.__class__.__name__}({self.name if self.name else ''}): {len(self.value)} entries " + "{\n"
        for tag in self.value:
            res += f"{tag.format_string(layer + 1)}\n"
        if layer:
            res += '  ' * layer
        res += "}"
        return res

    def __str__(self):
        return self.format_string()

    def __repr__(self):
        return self.format_string()

    def get(self, key, default=None):
        for tag in self.value:
            if tag.name == key:
                return tag
        return default

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value: TAG):
        if not isinstance(value, TAG):
            raise TypeError("value must be a TAG")
        for tag in self.value:
            if tag.name == key:
                value.name = key
                self.value[self.value.index(tag)] = value
                return
        value.name = key
        self.value.append(value)

    def __delitem__(self, key):
        for tag in self.value:
            if tag.name == key:
                self.value.remove(tag)
                return

    def __iter__(self):
        return iter(self.value)

    def __contains__(self, item):
        for tag in self.value:
            if tag.name == item:
                return True
        return False

    def keys(self):
        return [tag.name for tag in self.value]

    def values(self):
        return [tag.value for tag in self.value]

    def items(self):
        return [(tag.name, tag.value) for tag in self.value]

    def __len__(self):
        return len(self.value)


def get_tag_class(tag_id: TagId):
    TAGS = {
        TagId.TAG_END: TAG_End,
        TagId.TAG_BYTE: TAG_Byte,
        TagId.TAG_SHORT: TAG_Short,
        TagId.TAG_INT: TAG_Int,
        TagId.TAG_LONG: TAG_Long,
        TagId.TAG_FLOAT: TAG_Float,
        TagId.TAG_DOUBLE: TAG_Double,
        TagId.TAG_BYTE_ARRAY: TAG_Byte_Array,
        TagId.TAG_STRING: TAG_String,
        TagId.TAG_LIST: TAG_List,
        TagId.TAG_COMPOUND: TAG_Compound,
        TagId.TAG_INT_ARRAY: TAG_Int_Array,
        TagId.TAG_LONG_ARRAY: TAG_Long_Array,
    }
    res = TAGS.get(tag_id)
    if res is None:
        raise ValueError(f'Unknown tag id: {tag_id}')
    return res



from typing import Union

from .entities import *
from .snbt import SNBTParser


def parse_binary(binary: Union[BytesIO, bytes]) -> TAG_Compound:
    if isinstance(binary, bytes):
        binary = BytesIO(binary)

    root_tag_id = TagId(Struct('<b').unpack(binary.read(1))[0])
    return get_tag_class(root_tag_id)(binary=binary)


def parse_snbt(snbt: str) -> TAG_Compound:
    return SNBTParser(snbt).parse()
